# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

package Hits;
use lib '.';
use General;
use strict;

# headers are used for printing in the right order mainly...
our %headers = (
	       'PFAM-CONTEXT' => [qw(start end acc name GA start2 end2 score scoreHmm scoreContext E scoreSeq scoreHmmSeq scoreContextSeq ESeq mode)],
	       'PFAM' => [qw(start end acc name GA start2 end2 score E scoreSeq ESeq mode)],
	       'LCSPRE' => [qw(start end acc)],
	       );

sub deepCopyHits {
    # unfortunately, Standard Pfam and dPUC end up sharing hits, so modifying them on dPUC ends up modifying it on Standard Pfam too, and that's not good!
    # this detaches the hits so each of these methods can edit the hits without affecting each other
    my ($hits) = @_;
    my @hitsNew;
    foreach my $h (@$hits) {
	my %hNew = %$h; # this is the deepest level, things are effectively unlinked now
	push @hitsNew, \%hNew;
    }
    return \@hitsNew;
}

sub oneBasedEndInc {
    # need to make hits 1-based, end inclusive (non-computer people prefer that)
    # edits and returns copies, originals are unchanged...
    my ($hits) = @_;
    my @hitsNew;
    foreach my $h (@$hits) {
	my %hNew = %$h;
	# these two lines make system 1-based, end inclusive
	$hNew{start}++;
	$hNew{start2}++;
	# put in new list
	push @hitsNew, \%hNew;
    }
    return \@hitsNew;
}

sub filterHitsBooleanCol {
    # is a generalization of filtering using any of the Pfam cutoffs we have precalculated into the columns (GA,TC,NC).
    # if the file is missing the required column, nothing passes
    my ($hits, $col) = @_; # col is GA,TC,NC, or something else that can be treated as a boolean
    my @hitsNew;
    foreach my $h (@$hits) {
	push @hitsNew, $h if $h->{$col};
    }
    return \@hitsNew;
}

sub filterPfamClans {
    # for removing overlapping hits of the same clan, keeps only the one with the best score
    # can also be used for Superfamily to remove overlapping hits of the same fold, provided the right mapping is passed in $acc2clan
    
    # Score is defined by $col, and the order is expressed in the boolean $isDescending.
    # For example for the last two parameters, for bit scores we would use ('score', 1), for E-values ('E', 0).
    # remember not to use this on Superfamily hits, only PFAM
    # this script will not check that all entries are valid PFAM, that would make it slower (not sure if significantly so though)
    my ($hits, $acc2clan, $col, $isDescending) = @_;
    # defaults: PFAM (in pfam_scan.pl) uses E-values rather than scores to rank the families
    unless (defined $col) {
	$col = 'E';
	$isDescending = 0;
    }
    # sort hits by score
    my $hitsSorted = sortHitsByCol($hits, $col, 0, $isDescending); # assume score is a number (not such a bad assumption ;))
    my @hitsNew;
    foreach my $h (@$hitsSorted) {
	my $clan = $acc2clan->{$h->{acc}};
	unless (defined $clan) { push @hitsNew, $h; next; } # hit automatically passes if it doesn not belong to a clan
	# if passes, $h belongs to a clan
	my $isOverlapping = 0;
	foreach my $h2 (@hitsNew) {
	    my $clan2 = $acc2clan->{$h2->{acc}};
	    next unless defined $clan2; # proceed to comparison if this also belongs to a clan
	    if ( overlapRegionsObject($h, $h2) ) { # $h overlaps with a region that has already passed
		if ($clan eq $clan2) { $isOverlapping = 1; last; } # if clans are also the same, toss this $h
	    }
	}
	push @hitsNew, $h unless $isOverlapping; # will only pass if it didn't overlap with another domain of the same clan that ranked higher
    }
    return sortHitsByCol(\@hitsNew, 'start'); # sort for print now, in case this is the end of the line
}

sub sortHitsByCol {
    # a general method for sorting by the value of a column $col
    # $isText is a boolean, so default is to sort as numbers
    # $isDescending is also a boolean, so default is to sort ascending
    my ($hits, $col, $isText, $isDescending) = @_;
    return $hits unless defined $hits; # bail out early if hits are undefined (return input, no diff), so programs don't complain with missing entries
    my @hitsNew; # will hold the sorted hits
    if ($isText) {
	if ($isDescending) {
	    @hitsNew = sort { $b->{$col} cmp $a->{$col} } @$hits;
	}
	else {
	    @hitsNew = sort { $a->{$col} cmp $b->{$col} } @$hits;
	}
    }
    else {
	if ($isDescending) {
	    @hitsNew = sort { $b->{$col} <=> $a->{$col} } @$hits;
	}
	else {
	    @hitsNew = sort { $a->{$col} <=> $b->{$col} } @$hits;
	}
    }
    return \@hitsNew;
}

sub removeOverlapsByListOrder {
    # this sub takes a list of hits and removes all overlaps, keeping whatever came first in the list
    # the idea is that the hits have been sorted outside this sub with a particular priority in mind
    # we allow some overlaps via a nesting net, which maps accessions to the boolean of interest
    my ($hits, $nestingNet) = @_;
    my $haveNestingNet = defined $nestingNet && scalar keys %$nestingNet; # a boolean we use...
    my @hitsNew; # hits that pass
    foreach my $h (@$hits) {
	my $isOverlapping = 0;
	foreach my $h2 (@hitsNew) {
	    # $h overlaps with a region that has already passed
	    if ( overlapRegionsObject($h, $h2) ) {
		# overlap isn't allowed
		# if we don't have a nesting net, we won't bother looking for accessions
		unless ( $haveNestingNet && $nestingNet->{$h->{acc}}{$h2->{acc}} ) {
		    $isOverlapping = 1; last;
		}
	    }
	}
	# will only pass if it didn't overlap with another domain that ranked higher
	push @hitsNew, $h unless $isOverlapping;
    }
    return \@hitsNew;
}

sub removeOverlapsByScore {
    # I tried to be more general than Pfam.  This removes all overlaps, not only the ones of the same accession, unless they're allowed as defined by a hash.  The priority is set by the E-value (essentially a normalized score).  This should be useful for Supfam and Smart processing as well!!!  Again, this is more general than that processing, since it assumes no GA filtering (can do with predictions in general), and the nestingNet allows special overlaps (a feature that only Pfam has, in a way, but may be defined for Supfam as well)
    my ($hits, $nestingNet) = @_;
    $hits = sortHitsByCol($hits, 'E'); # sort by "normalized score"
    $hits = removeOverlapsByListOrder($hits, $nestingNet); # remove overlaps by score
    return sortHitsByCol($hits, 'start'); # re-sort by start, return
}

sub removeOverlapsPfam {
    # this removes overlaps of the same accession (presumably one is "ls", the other "fs") by keeping the one that the Pfam database prefers (comes from hash file)
    # sometimes this criterion is by score (actually E-value, as is done in current pfam_scan.pl script), sometimes it just prefers the longest or the shortest one, as long as both passed
    my ($hits, $acc2order) = @_;
    
    # I think the easiest thing is to go through each accession type and work on that subsets separately
    my %acc2hits;
    foreach my $h (@$hits) {
	my $acc = $h->{acc};
	push @{$acc2hits{$acc}}, $h; # add to this array of hits with the same accession
    }
    # now go through accessions
    my @hitsNew; # will hold things that pass
    while (my ($acc, $hitsThisAcc) = each %acc2hits) {
	# first sort things according to the order preferences of the accession (first will stay over second): (globalfirst, localfirst, byscore)
	my $order = $acc2order->{$acc}; # this is the order preference for this accession
	# these three values are going to be set by the method chosen to sort with
	my @sortParams; # ($col, $isText, $isDescending)
	if ($order eq 'globalfirst') { @sortParams = ('mode', 1, 1); } # ls before fs, want to sort on column "mode" descending
	elsif ($order eq 'localfirst') { @sortParams = ('mode', 1, 0); } # fs before ls, want to sort on column "mode" ascending
	elsif ($order eq 'byscore') { @sortParams = ('E', 0, 0); } # low E-values first, want to sort on column "E" ascending
	# actually do the sorting
	$hitsThisAcc = sortHitsByCol($hitsThisAcc, @sortParams);
	# now remove overlaps as we go down this list
	$hitsThisAcc = removeOverlapsByListOrder($hitsThisAcc);
	# finally, we can add this list to the main list
	push @hitsNew, @$hitsThisAcc;
    }
    # done removing overlaps
    # sort by start position of all hits, in case this has to be printed out, then return
    return sortHitsByCol(\@hitsNew, 'start'); # sorted by start, default sort numbers and ascending
}

sub processHitsAsPfam {
    # keeps GA only, merges ls and fs and keeps one based on family order preferences, and lastly filters clans
    my ($hitsLs, $hitsFs, $acc2order, $acc2clan) = @_;
    # first we eliminate everything that didn't pass GA
    my $hitsLsNew = filterHitsBooleanCol($hitsLs, 'GA');
    my $hitsFsNew = filterHitsBooleanCol($hitsFs, 'GA');
    # now recalculate sequence scores (since set has been reduced)
    recalcScoreSeq($hitsLsNew);
    recalcScoreSeq($hitsFsNew);
    # then we merge results (stored in first structure, not sorted, but following operation will sort)
    push @$hitsLsNew, @$hitsFsNew;
    # remove overlaps based on family order preferences
    $hitsLsNew = removeOverlapsPfam($hitsLsNew, $acc2order);
    # lastly, filter by clans
    $hitsLsNew = filterPfamClans($hitsLsNew, $acc2clan);
    # done, return new Hits
    return $hitsLsNew;
}

sub recalcScoreSeq {
    # if we eliminated sequences by GA or other methods, the scoreSeq (sum of the scores of all the domain in the same family, HMMER calls it the "sequence" score) might be larger than it's true value because it's also counting eliminated domains
    # this method overwrites the sequence scores in place
    my $hits = shift;
    my %acc2scoreSeq;
    # we have to loop once to get the sum of scores per family (accession)
    foreach my $h (@$hits) {
	$acc2scoreSeq{$h->{acc}} += $h->{score};
    }
    # now we loop again to store these sequence scores
    foreach my $h (@$hits) {
	$h->{scoreSeq} = sprintf '%0.1f', $acc2scoreSeq{$h->{acc}};
    }
    # done, nothing to return
}

sub overlapRegionsObject {
    # for my lazy writing, an object version of "sub overlapRegions" below
    my ($o1, $o2) = @_;
    return overlapRegions($o1->{start}, $o1->{end}, $o2->{start}, $o2->{end});
}

sub overlapRegions {
    # algorithm based on the one from pfam_scan.pl, package HMMUnit (from the PFAM guys)
    my ($s1, $e1, $s2, $e2) = @_; # start1, end1, start2, end2
    ($s1, $e1, $s2, $e2) = ($s2, $e2, $s1, $e1) if $s1 > $s2; # force regions to be ordered so that $s1 <= $s2
    return $s2 < $e1; # not <= cause ends are exclusive (a la perl, but not like hmmer)
}

sub readHitsManyProts_OnTheFly_StartHeader {
    # unlike the regular "sub readHitsManyProts", this one reads hits as we go, instead of waiting until we have a complete set of proteins.
    # this *_StartHeader opens the filehandle and gets the header, then reads the first protein name (but not the rest of the entry) and returns all
    my ($fi) = @_;
    my $fhi = General::getInFh($fi);
    # header parsing
    my $header = <$fhi>;
    chomp $header;
    my @header = split /\t/, $header;
    # first protein name
    my $prot = <$fhi>;
    if (defined $prot) { # handles the very rare case of a whole file with no protein entries... returns undef for prot
	chomp $prot;
	$prot =~ s/^\>//;
    }
    # done, return all
    return ($fhi, \@header, $prot);
}

sub readHitsManyProts_OnTheFly {
    # OnTheFly doesn't do GA, nor does it compute coverage
    # protNext will be undefined when last protein was reached
    my ($fhi, $header, $prot) = @_;
    my @header = @$header; # dereference header, makes parsing faster
    my $protNext;
    my @hits;
    while (<$fhi>) {
	chomp;
	if (s/^\>//) { $protNext = $_; last; } # found a new protein
	else {
	    # parsing an entry for this same protein
	    my %data;
	    @data{@header} = split /\t/;
	    push @hits, \%data;
	}
    }
    return ($prot, \@hits, $protNext);
}

sub printHitsManyProts_OnTheFly_StartHeader {
    # unlike the regular "sub printHitsManyProts", this one can print hits as we go, instead of waiting until we have a complete set of proteins.
    # this *_StartHeader version opens the filehandle, and gets and prints the header, then returns both filehandle and header.  The filehandle and header are used by the _OnTheFly version
    # note filehandle has to be closed outside
    my ($fo, $format, $comp) = @_;
    my $header = $headers{$format} || die "Error: a header was not found for format '$format'!";
    my $fho = General::getOutFh($fo, $comp); # cannot append if using compression
    print $fho join("\t", @$header)."\n"; # print header, no need to check if it's already there since only this process is writing to it
    return ($fho, $header);
}

sub printHitsManyProts_OnTheFly {
    # given an open filehandle and header already retrieved, prints this batch of hits
    # note filehandle has to be closed outside
    my ($fho, $header, $prot2hits, $prots) = @_;
    my @header = @$header; # dereference header, makes printing faster
    $prots = [keys %$prot2hits] unless defined $prots; # default, if no protein list was provided just print then in a random order
    foreach my $p (@$prots) { # to print proteins in same order as alignment, and so models make sense
	my $hits = $prot2hits->{$p};
	next unless $hits && @$hits;
	print $fho '>'.$p."\n"; # print protein name
	foreach my $h (@$hits) {
	    print $fho join("\t", @{$h}{@header})."\n"; # print hit, complicated slice
	}
    }
}

sub parseHmmerGeneral {
    # I've implemented this so it can extract all the data out of a run regardless of the database used (Pfam, Supfam, Smart).  To make sense out of these hits, post-processing might be required on the output data structure.
    # input may be either file path or the text content of that file (useful for my "piping" code)
    
    # takes hmmer output file (or its content in a string) and turns it into a Hits array structure with all the hit hashes
    # assumes we used the --acc option (it prints names otherwise and this system doesn't handle those).  Actually doesn't matter as long as the user outside knows what to do with the acc/name.  It always gets parsed correctly cause accs and names aren't allowed to have spaces.
    # hits are automatically ordered by starting position, then ending position, because this is how HMMer outputs it
    
    my ($fi) = @_;
    
    # decide whether input is file or the text content
    my @lines;
    if ($fi =~ /\n/) { # if input has newlines, it has to be text (filepaths shouldn't have newlines)
	@lines = split "\n", $fi; # use newlines to split whole text into lines.  end result will have no newlines in the end to worry about...
    }
    else { # if (General::fileExists($fi)) { # file exists
	my $fhi = General::getInFh($fi);
	@lines = <$fhi>;
	close $fhi;
	chomp @lines; # remove all newlines
    }
    
    my @hits;
    my $line = ''; # stores current line
    # map individual domains into sequence scores for this sequence
    my %acc2scoreSeq;
    my %acc2ESeq;
    
    # read protein sequence name (which we input in fasta format anyway)
    my $prot;
    until (($prot) = $line =~ /Query sequence:\s+(\S+)$/) { $line = shift @lines; } # read next line
    
    # read ahead until just before the first sequence hit appears
    until ($line =~ /^-/) { $line = shift @lines; } # read next line
    
    $line = shift @lines; # first hit
    while ($line && $line !~ /\[no hits above thresholds\]/) {
	# parse line
	# we can't just split by whitespace here (cause Description has spaces), it sucks...
	my ($acc, $scoreSeq, $ESeq) = $line =~ /^(\S+) +.+ +(-?\d+\.\d) +([\d\.e+-]+) +\d+$/;
#	die "Error: file $fi, this line could not be parsed!\n$line\n..." unless defined $acc;
	$acc2scoreSeq{$acc} = $scoreSeq;
	$acc2ESeq{$acc} = $ESeq;
	# read next line
	$line = shift @lines;
    }
    
    # read ahead until just before the first domain hit appears
    until ($line =~ /^-/) { $line = shift @lines; } # read next line
    
    $line = shift @lines; # first hit
    while ($line && $line ne '//' && $line !~ /\[no hits above thresholds\]/) {
	# parse line
	# things in "undef"s are: domain number (1/1), protein completeness cartoon (.]), domain completeness cartoon
	my ($acc, undef, $start, $end, undef, $start2, $end2, undef, $score, $E) = split / +/, $line;
#	die "Error: file $fi, this line could not be parsed!\n$line\n..." unless defined $acc;
	
	# alter starts so we have a 0-based end exclusive position system
	$start--;
	$start2--;
	
	# add to hits
	push @hits, {
	    'start' => $start,
	    'end' => $end,
	    'acc' => $acc, 
	    'start2' => $start2,
	    'end2' => $end2,
	    'score' => $score,
	    'E' => $E,
	    'scoreSeq' => $acc2scoreSeq{$acc}, # i.e. score2
	    'ESeq' => $acc2ESeq{$acc},
	};
	
	# read next line
	$line = shift @lines;
    }
    
    # finally done, return
    return ($prot, \@hits);
    
    # a sample generated with this command:
    # % ./hmmpfam --xnu --acc --forward -A 0 ~/databases/pfam/23/Pfam_ls test.fa
    # NOTE: why are the sequence and domain scores different below (when there's a single domain)???  This seems to be a consequence of the Forward algorithm.  Actually HMMer's commented C code explains this well, the discrepancy is noted and no good solutions are proposed.  We'll simply not use --forward for now.  HMMer 3 should deal with this a lot better, since they propose using --forward scores as a default anyway.
    
    # hmmpfam - search one or more sequences against HMM database
    # HMMER 2.3.2 (Oct 2003)
    # Copyright (C) 1992-2003 HHMI/Washington University School of Medicine
    # Freely distributed under the GNU General Public License (GPL)
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # HMM file:                 /home/ochoa/databases/pfam/23/Pfam_ls
    # Sequence file:            test.fa
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # 
    # Query sequence: PF10_0273
    # Accession:      [none]
    # Description:    [none]
    # 
    # Scores for sequence family classification (score includes all domains):
    # Model           Description                             Score    E-value  N
    # --------        -----------                             -----    ------- ---
    # PF01529.12      DHHC zinc finger domain                  87.1    6.2e-23   1
    # 
    # Parsed for domains:
    # Model           Domain  seq-f seq-t    hmm-f hmm-t      score  E-value
    # --------        ------- ----- -----    ----- -----      -----  -------
    # PF01529.12        1/1     102   166 ..     1    66 []    86.0  1.4e-22
    # //
}

sub parseHmmerPfam {
    # wrapper around "parseHmmerGeneral" that incorporates typical Pfam processing to give us the final structure we want
    # to keep it general, need to let the user specify which mode (ls or fs) these hits are in
    # as with parseHmmerGeneral, input may be either file path or the text content of that file (useful for my "piping" code)
    my ($fi, $acc2name, $acc2ds2ga, $mode) = @_;
    # parse using general HMMer parser
    my ($prot, $hits) = parseHmmerGeneral($fi);
    # edit hits in place instead of making copies
    foreach my $h (@$hits) {
	# get this accession
	my $acc = $h->{acc};
	# remove version number for accession, write back to hash
	$acc =~ s/\.\d+$//;
	$h->{acc} = $acc;
	# write name of family
	$h->{name} = $acc2name->{$acc};
	# calculate whether this hit passes GA or not (ignores overlaps from clans and modes, doesn't matter to us at this point)
	my $ds2ga = $acc2ds2ga->{$acc}; # copy down structure
	$h->{GA} = ($h->{scoreSeq} >= $ds2ga->{s} && $h->{score} >= $ds2ga->{d}) ? 1 : 0; # passed gathering cutoff or not (boolean)
	# write mode
	$h->{mode} = $mode;
    }
    # done, return processed structure
    return ($prot, $hits);
}

sub getPfamThresholds {
    # reads parsed PFAM thresholds, turns into a structure that is more manageable than just the plain hash file...
    # input GA thresholds parsed file path
    my $fi = shift;
    # get plain hash
    my $acc2thresholds = General::readHashFile($fi);
    # want to return these different structures
    my %mode2acc2ds2thresholds;
    my %acc2mode2ds2thresholds;
    while (my ($acc,$t) = each %$acc2thresholds) {
	# according to PFAM documentation, this is the order of the fields:
	# ls sequence, ls domain, fs sequence, fs domain.
	# sample line looks like this:
	# 8.60 8.60; 25.00 25.00;
	my ($t_ls, $t_fs) = split /; */, $t;
	my ($ls_s, $ls_d) = split / +/, $t_ls;
	my ($fs_s, $fs_d) = split / +/, $t_fs;
	# force sequence thresholds to be at least as high as the domain thresholds (an assumption that has no effect on setting thresholds but affects other analyses)
	$ls_s = $ls_d if $ls_s < $ls_d;
	$fs_s = $fs_d if $fs_s < $fs_d;
	# little structures
	my $sd2tLs = {
	    's' => $ls_s,
	    'd' => $ls_d,
	};
	my $sd2tFs = {
	    's' => $fs_s,
	    'd' => $fs_d,
	};
	# place in big structures
	$mode2acc2ds2thresholds{ls}{$acc} = $sd2tLs;
	$mode2acc2ds2thresholds{fs}{$acc} = $sd2tFs;
	$acc2mode2ds2thresholds{$acc} = {
	    'ls' => $sd2tLs,
	    'fs' => $sd2tFs,
	};
    }
    # return new hashes
    return (\%mode2acc2ds2thresholds, \%acc2mode2ds2thresholds);
}

sub getPfamThresholdsSeqBol {
    # this function makes a map for each accession and mode to a boolean that tells us whether applying a sequence threshold is necessary (since often it is redundant with the domain threshold).  Precomputing this saves us a lot of time!
    # Things are faster if we only save the accessions that had any true booleans (but we'll have to check before attempting to retrieve the modes)
    # input are the pfam thresholds
    my ($acc2mode2ds2t) = @_;
    my %acc2mode2seqBol;
    while (my ($acc, $mode2ds2t) = each %$acc2mode2ds2t) {
	my %mode2seqBol;
	while (my ($mode, $ds2t) = each %$mode2ds2t) {
	    # this sequence threshold has to be applied, since it's higher than the domain threshold (so it's not trivially satisfied)
	    $mode2seqBol{$mode} = 1 if $ds2t->{s} > $ds2t->{d};
	}
	$acc2mode2seqBol{$acc} = \%mode2seqBol if scalar keys %mode2seqBol;
    }
    return \%acc2mode2seqBol;
}

sub getNestingNet {
    # some accessions are allowed to overlap, since they are actually nested.  Pfam keeps a list of the cases that are documented, we use them to allow these overlaps in our lp_solve code that removes all other overlaps.
    # makes a net both ways, for simplicity
    # input is Pfam parsed nesting net hash file path
    my $fi = shift;
    # get plain hash first
    my $acc2nesting = General::readHashFile($fi);
    my %net;
    # navigate keys
    while (my ($acc1, $partners) = each %$acc2nesting) {
	my @acc2 = split /;/, $partners; # this line may have multiple accessions, separate by semicolon
	foreach my $acc2 (@acc2) {
	    # add to net both ways
	    $net{$acc1}{$acc2} = 1;
	    $net{$acc2}{$acc1} = 1;
	}
    }
    return \%net;
}

sub overlapNetBinarySearch {
    # this calculates an overlap networt (both ways) from scratch, doesn't assume anything about the starting data, and shouldn't alter it either
    # if available, uses an "allowed" net (like the Pfam nesting net) to omit overlaps from net.  Right now the net has to be exactly the Pfam nesting net because it uses accessions rather than hit references to do the mapping!!!  It's not worth doing the mapping outside in a different structure, the map is a one-time thing.  Keep in mind for future applications though!
    # uses the binary search subs to make this very fast
    my ($hits, $nestingNet) = @_;
    
    # at least once a protein sent an empty hits array.  Let's just return an empty net in that case (it should not be used anyway)
    my $numHits = scalar @$hits;
    return {} unless $numHits;
    
    my %overlapNet; # the structure we want
    my $haveNestingNet = defined $nestingNet; # a boolean that tells us quickly if we have a net or not
    
    # initialize the data (sorting and lowStart mapping)
    ($hits, my $hit2lowStart) = overlapBinarySearchInit($hits);
    
    # now start searching, ascending in array, and avoid rechecking old overlaps by adjusting the 'lower' range
    my $upper = $numHits-1; # precalculate the upper range, should speed things up a bit
    for (my $i = 0; $i < $numHits-1; $i++) { # note the last hit is never the query, since it's always in other query's results
	my $hi = $hits->[$i]; # the query
	my $lower = $i+1; # this skips the query and the things before it
	# these are the things that overlap our query
	my $hitsOv = overlapBinarySearch($hi, $hits, $hit2lowStart, $lower, $upper);
	# now we need to add the results to the net, both ways
	my $nestingNetHere = $nestingNet->{$hi->{acc}} if $haveNestingNet; # copy down structure to reduce hash calls
	foreach my $hj (@$hitsOv) {
	    # skip if overlap is allowed
	    next if $haveNestingNet && $nestingNetHere->{$hj->{acc}};
	    $overlapNet{$hi}{$hj} = 1;
	    $overlapNet{$hj}{$hi} = 1;
	}
    }
    
    # done, return the network!
    return \%overlapNet;
}

sub overlapBinarySearchInit {
    # sorts the list the way we need it for queries, and also gets the hit2lowStart auxiliary array.  Doesn't alter the original array
    my ($hits) = @_;
    
    # at least once, it happened that a protein sent an empty hits array, but the initial $lowStart chokes on that!  Let's just return empty arrays instead of attempting to make things work
    my $numHits = scalar @$hits;
    return ([], []) unless $numHits;
    
    # the very first thing to do is to sort the hits by end first, then by start to break ties
    my @hits = sort { $a->{end} <=> $b->{end} || $a->{start} <=> $b->{start} } @$hits;
    
    # now we create the array that tells us what is the lowest start position remaining relative to the current position
    my @hit2lowStart;
    my $lowStart = $hits[$numHits-1]{start}; # the last entry initializes this variable
    # navigate it backwards
    for (my $i = $numHits-1; $i >= 0; $i--) {
	my $starti = $hits[$i]{start}; # get this start
	$lowStart = $starti if $lowStart > $starti; # update the current lowest start
	$hit2lowStart[$i] = $lowStart;
    }

    return (\@hits, \@hit2lowStart);
}

sub overlapBinarySearch {
    # this is a very efficient search for overlaps, using a binary-search strategy
    # it assumes the query list is already sorted by end, then breaking ties by start (both ascending)
    # we also definitely need an array that for each position marks the lowest remaining start position
    # lower and upper are not required, but can be set to reduce searching ranges (if we've traversed the array up to down, for example).  They are in terms of the array indeces, not in terms of start or end coordinates!
    
    # binary search code based on General::binarySearch, itself based on/stolen from code: http://staff.washington.edu/jon/dsa-perl/bsearch
    # the rest of the algorithm was informed by http://stackoverflow.com/questions/303591/a-range-intersection-algorithm-better-than-on
    # particularly, this entry:
    #   1. Use binary search to find the first range with an End value of >= TestRange.Start
    #   2. Iterator starting at the binary search until you find an Start > TestRange.End:
    #      2a. If the range if the current range is within the TestRange, add it to your result.
    # I realized I needed to edit the structure to add one more array, one that maps each position to the lowest start position remaining (can be built in O(n), not bad compared to sorting)
    
    my ($hitQuery, $hits, $hit2lowStart, $lower, $upperIni) = @_;
    # the query hit itself matters very little compared to just knowing the start and end, so we extract them now
    my ($startQuery, $endQuery) = @{$hitQuery}{qw(start end)};
    
    # the first part is using a binary search to find the first range with end > query.start (this accounts for end-exclusive, the way we have it)
    # set these two default values
    $lower = 0 unless defined $lower;
    $upperIni = scalar(@$hits)-1 unless defined $upperIni;
    my $upper = $upperIni; # we'll remember what the initial upper value was, since $upper gets updated as the search progresses
    my $i; # index of probe
    while ($lower < $upper) {
	$i = int(($lower + $upper)/2); # choose the index right at the middle of our range
	my $end = $hits->[$i]{end}; # get the current end (remember this is end-exclusive)
	if ($end > $startQuery) { $upper = $i; } # this means we might have it, but we might need to look a bit lower
	else { $lower = $i+1; } # we definitely need to look higher.  The +1 guarantees that rounding down won't get us stuck in infinite iterations, right?
    }
    # since we're not looking for an exact number (many ranges can have the same end point, and we could have "hits" with identical ranges), we're only done when the range has been reduced to a point
    if ($lower == $upper) { $i = $lower; } # this is our starting point
    elsif ($upper == $upperIni && $lower == $upper+1) { return []; } # this is an anomalous case, happens when our query's start was higher than any of the ends present, so no hits were found!
    else { die "Unexpected error: lower $lower != upper $upper (upperIni $upperIni)!"; } # I don't think this ever happens
    
    # the second part is to iterate up until we find a hit with lowStart >= query.end (accounts for end-exclusive setup again)
    # we need to *check* and collect hits until this happens
    my @hitsOv;
    # we get the first start to start the loop (this also verifies that we got anything at all, failure to satisfy the condition means nothing overlapped our hit)
    my $lowStart = $hit2lowStart->[$i];
    while ($lowStart < $endQuery) {
	# this hit might be promising, but we need to check it too
	my $hi = $hits->[$i];
	push @hitsOv, $hi if $hi->{start} < $endQuery; # this hit was good
	# look at next hit
	$i++;
	last if $i > $upperIni; # check that we don't go over bounds (happens if the very last range overlapped with our query)
	$lowStart = $hit2lowStart->[$i];
    }
    # done! return 
    return \@hitsOv;
}

# for posElim only, but could be used by nonOvNeg if we wanted to skip posElim altogether
sub resolveLsFs {
    # this decides what stays of ls-fs for a single protein, given the score of each part and the thresholds
    # we keep ls if H_i^ls-T_i^ls > sum_j H_j^fs-T_j^fs, otherwise we keep the fs (all the fragments if there's many)
    # things that are not paired ls-fs automatically stay
    my ($hitsLs, $hitsFs, $acc2mode2ds2t) = @_;
    
    # handle trivial cases...
    my $numHitsLs = defined $hitsLs ? scalar @$hitsLs : 0;
    my $numHitsFs = defined $hitsFs ? scalar @$hitsFs : 0;
    # in this case we had no hits, so the list will be empty
    unless ($numHitsLs + $numHitsFs) { return []; }
    # in this case we have at least one hit of some kind
    else {
	# if one of the lists is empty, return the other one (since we know there's at least one something)
	return $hitsFs unless $numHitsLs;
	return $hitsLs unless $numHitsFs;
	# else both lists have hits, continue to more complicated code...
    }
    
    my @hitsPass; # the things that will go on to the next round
    my @hitsFsToTest; # the things that we need to test specifically among the fs (because they overlapped with a related ls).  If we separate them first, it will get faster to do the precise sum-of-scores evaluation later (when we're looking at the ls side)
    
    # we hash ls hits by their accession, should make the search below much faster when the fs list is very large (which it is frequently)
    my %acc2hitsLs;
    foreach my $hLs (@$hitsLs) {
	push @{$acc2hitsLs{$hLs->{acc}}}, $hLs;
    }
    foreach my $hFs (@$hitsFs) {
	# figure out if they overlap something with the same accession in the ls list
	# instead of going down the list of ls hits and asking one by one if the accession matches, we decided to hash them by their accession, so the question then becomes automatically answered.
	my $overlapLs = 0;
	my $hitsLsThisAcc = $acc2hitsLs{$hFs->{acc}}; # get list of interest
	if (ref $hitsLsThisAcc) { # attempt the loop only if this list exists (undefined value will make perl die)
	    # factor out retrieval of start,end from hash (to reduce hash calls)
	    my $startFs = $hFs->{start};
	    my $endFs = $hFs->{end};
	    foreach my $hLs (@$hitsLsThisAcc) {
		if ( overlapRegions($startFs, $endFs, $hLs->{start}, $hLs->{end}) ) {
		    # since they have the same accession and they overlap, then this will have to be tested later
		    $overlapLs = 1;
		    last;
		}
	    }
	}
	# decide the fate of this fs prediction
	if ($overlapLs) { push @hitsFsToTest, $hFs; }
	else { push @hitsPass, $hFs; }
    }
    # we do the hashing of fs hits by their accession as well, should make the search below much faster when the ls list is very large
    my %acc2hitsFsToTest;
    foreach my $hFs (@hitsFsToTest) {
	push @{$acc2hitsFsToTest{$hFs->{acc}}}, $hFs;
    }
    # now we should have a small list of fs domains that have to be tested, so we'll go through each ls and decide whether it or the fs ones pass
    foreach my $hLs (@$hitsLs) {
	# collect the list of fs hits that overlap it (sometimes it's 0 or 1, but it could be multiple)
	my @hitsFsOverlapping;
	my $hitsFsToTestThisAcc = $acc2hitsFsToTest{$hLs->{acc}}; # get list of interest
	if (ref $hitsFsToTestThisAcc) { # attempt the loop only if this list exists (undefined value will make perl die)
	    # factor out retrieval of start,end from hash (to reduce hash calls)
	    my $startLs = $hLs->{start};
	    my $endLs = $hLs->{end};
	    foreach my $hFs (@$hitsFsToTestThisAcc) {
		if ( overlapRegions($startLs, $endLs, $hFs->{start}, $hFs->{end}) ) {
		    push @hitsFsOverlapping, $hFs;
		}
	    }
	}
	# if zero, then this ls hit automatically passes
	unless (@hitsFsOverlapping) { push @hitsPass, $hLs; }
	# else we'll have to compute the two normalize sum of scores to decide which set stays
	else {
	    # get thresholds once
	    my $mode2ds2t = $acc2mode2ds2t->{$hLs->{acc}};
	    my $tLs = $mode2ds2t->{ls}{d};
	    my $tFs = $mode2ds2t->{fs}{d};
	    # calculate both scores
	    # there can only be one ls, but there may be many fs
	    my $scoreLs = $hLs->{score} - $tLs;
	    my $sumScoresFs = 0;
	    foreach my $hFs (@hitsFsOverlapping) {
		$sumScoresFs += $hFs->{score} - $tFs;
	    }
	    # compare them and decide which domains pass then
	    if ($scoreLs >= $sumScoresFs) { push @hitsPass, $hLs; } # detail, also keep ls if they're tied!!! (unclear what would be best)
	    else { push @hitsPass, @hitsFsOverlapping; }
	}
    }
    # completely done, now we only have one set for each ls-fs conflict we used to have
    return \@hitsPass;
}

1;
