# Copyright (C) 2010 Alejandro Ochoa, Singh Research Group, Princeton University
# This file is part of dPUC.
# dPUC is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# dPUC is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with dPUC.  If not, see <http://www.gnu.org/licenses/>.

use lib '.';
use General;
use FilePaths;
use Hits;
use strict;

# this sub parses and organizes all the Pfam hits defined in Pfam-A.full so that we have our familiar Hits.pm file sorted first by protein, then by hits with start and end coordinates, sorted by start.  We will extract context from this file, with filtering options that are too complex to be part of this script.  This just takes the hits from Pfam and puts them in a nice file of ours.

# a big problem with parsing and storing data naively is memory.  Our regular Hits.pm structure takes up too much memory as it is, so we decided not to start by storing things with this structure.  Therefore, a big part of our parsing strategy is to store hits as text with each hit concatenated in a random order and with the Pfam accession, start, and end built into the string in an easily-parseable way.  This way we avoid all the memory overhead associated with Perl structures, since each protein has only one very long string (no extra overhead per hit or per hit datum).  In the end, when parsing is done, we can navigate each protein and deflate the string into the structures we like so we can sort by start and print with familiar code.  We can afford to do this only a protein at the time, so we print on the fly.

# the parsing code was originally based on parsePfamASeed.pl, but has diverged out of necessity

# constants
my $format = 'LCSPRE'; # format coincides with LCSPRE, but this is not a pre-LCS!!!
my $comp = 'gzip'; # compress output Hits file

# some data to initialize
my %prot2hits; # build predictions per protein

# temporary variables
my $acc; # current domain we're parsing

# load and read the Pfam-A.full file
print "Parsing Pfam-A.full...\n";
my $fhi = General::getInFh($FilePaths::pfamAFull);
while (<$fhi>) {
    # stick to old-school parsing for this one, might get more speed and modularity is not an issue for such a short and hacky script
    if (/^\#=GF AC +PF(\d{5})/) { # note this won't keep the Pfam version, or the initial PF, we don't care about them (will save us a bit of memory)
	$acc = $1;
    }
    elsif (/^\#=GS \w+\/(\d+)\-(\d+) +AC (\w+)/) {
	# kept var names for reference, but they're copied directly where they go so we don't waste time creating/copying vars just to not really use them
#	my $start = $1;
#	my $end = $2;
#	my $prot = $3; # get the accession, not the Uniprot name (since we want a non-redundant ID without the organism tag).  Note the regexp above doesn't save version number, we don't care about that either
	# process for hit list
	# compacted text version, should take a lot less memory (because Perl structures and even lots of scalars take up too much memory, but a very long string doesn't have much overhead)
	$prot2hits{$3} .= '/'.join '-',$acc,$1,$2; # don't copy parsed numbers to vars, just put them here directly to speed things up
    }
}
close $fhi;

print "Printing hits out...\n";
# to save memory, we'll need to print on the fly (we only process a protein at the time, we can't afford to do them all at once)
# initialize output file
my ($fho, $header) = Hits::printHitsManyProts_OnTheFly_StartHeader($FilePaths::pfamProt2hits, $format, $comp);
while (my ($prot, $hits) = each %prot2hits) {
    my @hits; # the final hits structure
    $hits =~ s/^\///; # the hits always start with one slash, remove it before splitting in the foreach loop below (that way we don't get empty hits)
    foreach my $hitTxt (split /\//, $hits) { # split by forward-slash
	my ($acc,$start,$end) = split /-/, $hitTxt; # split by -
	# store as structure now
	push @hits, {
	    'acc' => 'PF'.$acc, # note we only had the Pfam accession number before, here we add the PF prefix we had removed when parsing
	    'start' => $start - 1, # force 0-based, end-exclusive system
	    'end' => $end,
	};
    }
    # yey, got hits in structure
    # sort by start and save
    Hits::printHitsManyProts_OnTheFly($fho, $header, {$prot => Hits::sortHitsByCol(\@hits, 'start')}, undef, $comp);
}
close $fho;
